#include "DXUT.h"
#include "resource.h"
#include "App.h"
#include "Mesh.h"

#include "Vehicle.h"
#include "Ragdoll.h"
#include "BulletUtilities.h"
#include "ShapeDrawer.h"
#include "Utilities.h"
#include "HeightMap.h"

// Bullet includes
#include <btBulletDynamicsCommon.h>
#include <LinearMath/btQuickprof.h>
#include <BulletCollision/CollisionDispatch/btGhostObject.h>
#include <BulletCollision/CollisionShapes/btHeightfieldTerrainShape.h>
#include <BulletDynamics/Character/btKinematicCharacterController.h>

#include <list>

#define DRAW_STATIC_SHAPES
#define DRAW_SHAPES
//#define DRAW_VEHICLE_BOUNDINGSPHERE

#define HEIGHTMAP_WIDTH 100
#define HEIGHTMAP_HEIGHT 100

#ifdef DEBUG
const int gNumRigidBodies = 4;
const int gNumVehicles = 5;
const int gNumRagdolls = 5;
#else
const int gNumRigidBodies = 8;
const int gNumVehicles = 5;
const int gNumRagdolls = 10;
#endif

enum { IDC_TOGGLEPAUSESIMULATION = IDC_CLIENTCONTROLID, IDC_SHOOTSPEEDTEXT, IDC_SHOOTSPEED, IDC_JEDIFORCETEXT, 
IDC_JEDIFORCEIMPULSE, IDC_JEDIMODE, IDC_CYCLEVEHICLE, IDC_CHANGECAMERA, IDC_RESETSCENE };

class BulletRigidBodies : public App
{
protected:	
	void ResetScene()
	{
		std::list<btRigidBody*>::iterator list_iter;	
		for( list_iter = mShootRigidBodies.begin(); list_iter != mShootRigidBodies.end(); )
		{				
			btCollisionObject *obj = &**list_iter;
			list_iter++;

			btRigidBody* body = btRigidBody::upcast(obj);
			if (body && body->getMotionState())
			{
				delete body->getMotionState();
			}
			mDynamicsWorld->removeCollisionObject( obj );
			SAFE_DELETE( obj );			
		}

		mShootRigidBodies.clear();

		int numObjects = 0;
		if (mDynamicsWorld)
		{
			mDynamicsWorld->stepSimulation(1.f/60.f,0);
			numObjects = mDynamicsWorld->getNumCollisionObjects();
		}

		for (int i = 0; i < numObjects; i++)
		{
			btCollisionObject* colObj = mDynamicsWorld->getCollisionObjectArray()[i];
			btRigidBody* body = btRigidBody::upcast(colObj);
			if (body)
			{
				if (body->getMotionState())
				{
					btDefaultMotionState* myMotionState = (btDefaultMotionState*)body->getMotionState();
					myMotionState->m_graphicsWorldTrans = myMotionState->m_startWorldTrans;
					colObj->setWorldTransform( myMotionState->m_graphicsWorldTrans );
					colObj->setInterpolationWorldTransform( myMotionState->m_startWorldTrans );
					colObj->activate();					
				}
				// Removed cached contact points
				mDynamicsWorld->getBroadphase()->getOverlappingPairCache()->cleanProxyFromPairs(colObj->getBroadphaseHandle(), mDynamicsWorld->getDispatcher());

				btRigidBody* body = btRigidBody::upcast(colObj);
				if (body && !body->isStaticObject())
				{
					btRigidBody::upcast(colObj)->setLinearVelocity(btVector3(0,0,0));
					btRigidBody::upcast(colObj)->setAngularVelocity(btVector3(0,0,0));
				}
			}
		}
	}

	void InitBulletPhysics()
	{
		// Init physics and collision detection		
		mCollisionConfiguration = new btDefaultCollisionConfiguration();
		mCollisionDispatcher = new btCollisionDispatcher(mCollisionConfiguration);

		btVector3 worldMin( -10000.f, -10000.f, -10000.f );
		btVector3 worldMax( 10000.f, 10000.f, 10000.f );
		mBroadPhase = new btAxisSweep3(worldMin,worldMax,32766);
		//mBroadPhase = new btDbvtBroadphase();	
		mConstraintSolver = new btSequentialImpulseConstraintSolver();
		mDynamicsWorld = new btDiscreteDynamicsWorld( mCollisionDispatcher, mBroadPhase, mConstraintSolver, mCollisionConfiguration);
		mDynamicsWorld->getDispatchInfo().m_enableSPU = true;
		mDynamicsWorld->setGravity( btVector3(0, -9.8f, 0));	


	}	

	virtual HRESULT OneTimeClientInit()													
	{	
		mSceneMesh = new Mesh();
		mShapeDrawer = new ShapeDrawer();

		mHeightMap = new HeightMap( Point(HEIGHTMAP_WIDTH, HEIGHTMAP_HEIGHT), 20.f );		
		mHeightMap->CreateRandomHeightMap( rand() % 2000, 1.5f, 0.6f, 6 );		

		mAppDlg.AddButton( IDC_CHANGECAMERA, L"Change camera(F4)", 25, mCurrentGuiHeight += 24, 125, 22 );
		mAppDlg.AddButton( IDC_CYCLEVEHICLE, L"Cycle vehicle(F5)", 25, mCurrentGuiHeight += 24, 125, 22 );
		mAppDlg.AddButton( IDC_RESETSCENE, L"Reset scene(space)", 25, mCurrentGuiHeight += 24, 125, 22 );
		mAppDlg.AddStatic( IDC_SHOOTSPEEDTEXT, L"Shoot speed:", 25, mCurrentGuiHeight += 24, 125, 22    );
		mAppDlg.AddSlider( IDC_SHOOTSPEED, 25, mCurrentGuiHeight += 24, 125, 22, 0, 1000, (int)mShootSpeed );		
		mAppDlg.AddStatic( IDC_JEDIFORCETEXT, L"Jedi force strength:", 25, mCurrentGuiHeight += 24, 125, 22    );
		mAppDlg.AddSlider( IDC_JEDIFORCEIMPULSE, 25, mCurrentGuiHeight += 24, 125, 22, 0, 1000, (int)mShootSpeed );		
		mAppDlg.AddCheckBox( IDC_TOGGLEPAUSESIMULATION, L"Pause simulation (i)", 25, mCurrentGuiHeight += 24, 125, 22, false );	
		mAppDlg.AddCheckBox( IDC_JEDIMODE, L"Jedi mode", 25, mCurrentGuiHeight += 24, 125, 22, false );	

		InitBulletPhysics();

		return S_OK; 
	}

	virtual HRESULT CreateClient()
	{
		HRESULT hr = S_OK;

		Progress( "Creating scene...", .2f );
		mShapeDrawer->Create( mDevice );

		// Create a static plane and ghost object for jedi force
		{
			btTransform startTransform;
			startTransform.setIdentity ();		

			// Create a static plane for the ground
			btCollisionShape *groundShape = new btStaticPlaneShape(btVector3(0, 1, 0), 0);			
			mCollisionShapes.push_back(groundShape);	

			btRigidBody *groundBody = CreateRigidBody( .0, startTransform, groundShape );
			groundBody->setRestitution(0);
			groundBody->setFriction(1.0f);	
			mDynamicsWorld->addRigidBody(groundBody);	

			// Init ghost object
			btConvexShape *ghostObjectShape = new btBoxShape(btVector3(3.f, 3.f, 3.f));		
			mCollisionShapes.push_back( ghostObjectShape );

			startTransform.setOrigin (btVector3(0.0, 10000.0, 0.0));
			mGhostObject = new btGhostObject();
			mGhostObject->setCollisionShape(ghostObjectShape);
			mGhostObject->setWorldTransform(startTransform);
			mGhostObject->setCollisionFlags(btCollisionObject::CO_GHOST_OBJECT);	
			mGhostObject->setUserPointer( &mIsGhostObject );
			// Register callback or else our ghost object wouldn't register the overlaps
			mGhostPairCallback = new btGhostPairCallback();
			mBroadPhase->getOverlappingPairCache()->setInternalGhostPairCallback(mGhostPairCallback);
			mDynamicsWorld->addCollisionObject(mGhostObject, btBroadphaseProxy::CharacterFilter, btBroadphaseProxy::StaticFilter|btBroadphaseProxy::DefaultFilter);		

			// Create a sphere to shoot
			mShootShape = new btSphereShape( 1.0f );	
			mCollisionShapes.push_back( mShootShape );
		}

		// Create bunch of rigid bodies
		{
			const float sizexz = 1.f;		
			const float sizey = 2.f;		
			const float cubeSize = 1.f;	

			mCollisionShapes.push_back( CreatePyramid(mDynamicsWorld, btVector3(-20.0f,0.0f,-20.f),gNumRigidBodies/2,btVector3(sizexz,sizey,sizexz), CYLINDER_SHAPE_PROXYTYPE) );		
			mCollisionShapes.push_back( CreateWall(mDynamicsWorld, btVector3(-2.0f,0.0f,-20.f),gNumRigidBodies,btVector3(cubeSize,cubeSize,cubeSize)) );
			mCollisionShapes.push_back( CreateWall(mDynamicsWorld, btVector3(4.0f,0.0f,-20.f),gNumRigidBodies,btVector3(cubeSize,cubeSize,cubeSize)) );
			mCollisionShapes.push_back( CreateWall(mDynamicsWorld, btVector3(10.0f,0.0f,-20.f),gNumRigidBodies,btVector3(cubeSize,cubeSize,cubeSize)) );		
			mCollisionShapes.push_back( CreateTowerCircle(mDynamicsWorld, btVector3(25.0f,0.0f,-20.f),gNumRigidBodies,6,btVector3(cubeSize,cubeSize,cubeSize)) );	

			// Create rigid bodies on top of the terrain
			mCollisionShapes.push_back( CreatePyramid(mDynamicsWorld, btVector3(0.0f,20.0f,-120.f),gNumRigidBodies /2,btVector3(sizexz,sizey,sizexz), CYLINDER_SHAPE_PROXYTYPE) );		
		}

		// Init character controller
		{
			btTransform startTransform;
			startTransform.setIdentity ();
			startTransform.setOrigin (btVector3(0.0, 20.0, 0.0));

			// Looks like there's memory leak here...
			mCharacterGhostObject = new btPairCachingGhostObject();
			mCharacterGhostObject->setWorldTransform(startTransform);					

			btScalar characterHeight= 3.f;
			btScalar characterWidth = 1.f;
			btConvexShape* capsule = new btCapsuleShape(characterWidth,characterHeight);	
			mCollisionShapes.push_back(capsule);	
			mCharacterGhostObject->setCollisionShape (capsule);
			mCharacterGhostObject->setCollisionFlags (btCollisionObject::CF_CHARACTER_OBJECT);	

			btScalar stepHeight = btScalar(0.35);
			mCharacterController = new btKinematicCharacterController (mCharacterGhostObject, capsule, stepHeight);		

			mDynamicsWorld->addCollisionObject(mCharacterGhostObject,btBroadphaseProxy::CharacterFilter, btBroadphaseProxy::StaticFilter|btBroadphaseProxy::DefaultFilter);
			mDynamicsWorld->addCharacter(mCharacterController);				
		}	

		Progress( "Creating scene...", .4f );
		for( int i = 0; i < gNumVehicles; i++ )
		{
			D3DXVECTOR3 startPosition = RandomNumber( D3DXVECTOR3( -20.0f, 40.0f, -150.f ), D3DXVECTOR3( 50.0f, 60.0f, 50.f ) );
			Vehicle *vehicle = new Vehicle();
			vehicle->Create( mDevice, mDynamicsWorld, ConvertVector( startPosition ), 800.0f );
			mVehicles.push_back( vehicle );
		}		

		for( int i = 0; i < gNumRagdolls; i++ )
		{
			D3DXVECTOR3 startPosition = RandomNumber( D3DXVECTOR3( -40.0f, 40.0f, -150.f ), D3DXVECTOR3( 50.0f, 60.0f, 50.f ) );
			float size = RandomNumber( 3.f, 10.f );
			RagDoll *ragDoll = new RagDoll (mDynamicsWorld, ConvertVector( startPosition ), size );		
			mRagDolls.push_back( ragDoll );
		}		

		Progress( "Creating scene...", .6f );
		if(SUCCEEDED(mSceneMesh->Create( mDevice, "Media//scene.x" )))
		{
			Progress( "Creating scene...", .8f );

			mTriMeshData = new TriMeshData();		
			btConcaveShape *sceneShape = CreateBvhTriangleMeshShape(mSceneMesh->GetMesh(), mTriMeshData);
			mCollisionShapes.push_back(sceneShape);	
			sceneShape->setLocalScaling( btVector3( 1.5f, 1.5f, 1.5f ) );	
			//sceneShape->setMargin( 0.5f );

			btTransform trn;
			trn.setIdentity();	
			trn.setOrigin( btVector3( 0.f, 0.f, 0.f ) );
			btRigidBody *pScene = CreateRigidBody(0.0f,trn, sceneShape);
			mDynamicsWorld->addRigidBody( pScene );		

			mShapeDrawer->AddConcaveShape(sceneShape);		

		}		

		Progress( "Creating scene...", .9f );
		// Create a height field shape
		{
			bool flipQuadEdges = false;				
			btConcaveShape *heightFieldTerrainShape = new btHeightfieldTerrainShape(HEIGHTMAP_WIDTH, HEIGHTMAP_HEIGHT, mHeightMap->GetHeightTable(),
				1.0f, 0.f, 1000.f, 1, PHY_FLOAT, flipQuadEdges );		
			heightFieldTerrainShape->setLocalScaling( btVector3( 1.5f, 1.f, 1.5f ) );		
			mCollisionShapes.push_back(heightFieldTerrainShape);

			btTransform trn;
			trn.setIdentity ();
			trn.setOrigin (btVector3(0.0, -5.0, -125.0));	
			btRigidBody* heightMapBody = CreateRigidBody( .0, trn, heightFieldTerrainShape );
			heightMapBody->setRestitution(0);
			heightMapBody->setFriction(1.0f);	
			mDynamicsWorld->addRigidBody( heightMapBody );		
			mShapeDrawer->AddConcaveShape( (btConcaveShape*)heightFieldTerrainShape );	
		}


		return hr;
	}

	virtual HRESULT ResetClient()														
	{
		D3DXVECTOR3 vecEye( 0.0f, 0.0f, 0.0f );
		D3DXVECTOR3 vecAt ( 0.0f, 0.0f, 0.0f );
		mChaseCamera.SetProjParams( FOV, 1.33f, ZNEARPLANE, ZFARPLANE );
		mChaseCamera.SetViewParams( &vecEye, &vecAt );	
		mChaseCamera.SetEnablePositionMovement( true );		
		mChaseCamera.SetScalers( 0.01f, mCamSpeed );
		mChaseCamera.SetRotateButtons( false, false, true, false );			
		mCurrentCamera = &mCamera;

		mDevice->SetTransform( D3DTS_PROJECTION, mCurrentCamera->GetProjMatrix() );		

		return S_OK; 
	}

	virtual void LostClient()															
	{
	}

	virtual void DestroyClient()														
	{ 
		if( mShapeDrawer ) mShapeDrawer->Destroy();
		if( mSceneMesh ) mSceneMesh->Destroy();

		for(size_t i = 0; i < mRagDolls.size(); i++ )
		{
			SAFE_DELETE( mRagDolls[i] );
		}
		mRagDolls.clear();
		for(size_t i = 0; i < mVehicles.size(); i++ )
		{
			SAFE_DELETE( mVehicles[i] );
		}
		mVehicles.clear();

		// Cleanup in the reverse order of creation/initialization				
		if( mCharacterGhostObject )
		{
			mDynamicsWorld->removeCollisionObject( mCharacterGhostObject );		
			mDynamicsWorld->removeCharacter(mCharacterController);		
			SAFE_DELETE( mCharacterGhostObject );
			SAFE_DELETE( mCharacterController );	
		}

		if( mDynamicsWorld )
		{
			// Delete physics object
			for( int i = mDynamicsWorld->getNumCollisionObjects() - 1; i >= 0 ;i-- )
			{
				btCollisionObject* obj = mDynamicsWorld->getCollisionObjectArray()[i];
				btRigidBody* body = btRigidBody::upcast(obj);
				if (body && body->getMotionState())
				{
					delete body->getMotionState();
				}
				mDynamicsWorld->removeCollisionObject( obj );
				SAFE_DELETE( obj );			
			}

			// Delete collision shapes
			for (int j=0;j<mCollisionShapes.size();j++)
			{
				btCollisionShape* shape = mCollisionShapes[j];
				SAFE_DELETE( shape );			
			}
			mCollisionShapes.clear();
		}	

		SAFE_DELETE( mGhostPairCallback );		
		SAFE_DELETE( mTriMeshData );
	}

	virtual void RenderClient( double time, float elapsedTime )							
	{ 
		mDevice->SetTransform( D3DTS_VIEW, mCurrentCamera->GetViewMatrix() );		

		mShapeDrawer->ResetStats();

		// Set material for our concave shapes
		Material concaveShapeMtrl;
		concaveShapeMtrl.SetDiffuseColor( Color::Grey );
		concaveShapeMtrl.SetAmbientColor( Color::Grey );
		concaveShapeMtrl.SetSpecularColor( Color::Grey );
		mDevice->SetMaterial( &concaveShapeMtrl );

		// Render scene
		D3DXMATRIX matWorld;
		D3DXMatrixTranslation( &matWorld, 0.f, 0.f, 0.f );
		mShapeDrawer->RenderConcaveShapes( &matWorld, 0 );

		// Render terrain / height map
		D3DXMatrixTranslation( &matWorld, 0.f, -5.f, -125.f );
		mShapeDrawer->RenderConcaveShapes( &matWorld, 1 );

		// Render rigid bodies
		int objectDrawnCount = 0;
		if( mDynamicsWorld && mShapeDrawer )
		{
#ifdef DRAW_SHAPES
			Material shapeMaterial;			
			shapeMaterial.SetAmbientColor( 50, 50, 50 );
			shapeMaterial.SetSpecularColor( Color::White );
			shapeMaterial.SetPower( 50 );

			mDevice->SetRenderState(D3DRS_SPECULARMATERIALSOURCE, D3DMCS_MATERIAL);
			mDevice->SetRenderState( D3DRS_SPECULARENABLE, TRUE );

			const int numObjects = mDynamicsWorld->getNumCollisionObjects();

			// Draw everything within camera frustum
			for(int i =0; i < numObjects; i++)
			{
				btCollisionObject *colObj = mDynamicsWorld->getCollisionObjectArray()[i];			

				btVector3 center; btScalar radius;
				colObj->getCollisionShape()->getBoundingSphere( center, radius );
				center += colObj->getWorldTransform().getOrigin();

				if( mCurrentCamera->FrustumSphereCheck( ConvertVector(center), radius ) )
				{
					btRigidBody *body = btRigidBody::upcast(colObj);

					if(body && body->getMotionState())
					{	
#ifndef DRAW_STATIC_SHAPES
						if( body->getInvMass() == 0.0f )
							continue;
#endif
						// So the vehicle shape hull does not get drawn
						if( body->getCollisionShape()->getShapeType() <= CYLINDER_SHAPE_PROXYTYPE )
						{
							shapeMaterial.SetDiffuseColor( Color::Red );
							mDevice->SetMaterial( &shapeMaterial );

							btTransform trn = body->getWorldTransform();
							mShapeDrawer->Render( body->getCollisionShape(), trn );						
						}
					}
					else
					{
						shapeMaterial.SetDiffuseColor( Color::Blue );
						mDevice->SetMaterial( &shapeMaterial );

						// Don't draw the ghost object that triggers the jedi force, 
						// we get the user pointer we previously set for this
						bool isGhostObject = (bool*)colObj->getUserPointer();						
						if(!isGhostObject)
						{
							mShapeDrawer->Render( colObj->getCollisionShape(), colObj->getWorldTransform() );						
						}
					}
					objectDrawnCount++;
				}
			}
#endif	
		}
		// Set material for our vehicles
		Material vehicleShapeMtrl;
		vehicleShapeMtrl.SetDiffuseColor( Color::Blue );
		vehicleShapeMtrl.SetAmbientColor( 50, 50, 50 );
		vehicleShapeMtrl.SetSpecularColor( Color::White );
		vehicleShapeMtrl.SetPower( 50 );
		mDevice->SetMaterial( &vehicleShapeMtrl );

		// Render vehicles		
		for( size_t i = 0; i < mVehicles.size(); i++ )
		{
			D3DXMATRIX matWorld = ConvertMatrix( mVehicles[i]->GetWorldTransform() );		
			// Adjust center
			matWorld._41 += mVehicles[i]->GetMesh()->GetBoundingBox()->center.x;
			matWorld._42 += mVehicles[i]->GetMesh()->GetBoundingBox()->center.y;
			matWorld._43 += mVehicles[i]->GetMesh()->GetBoundingBox()->center.z;
			D3DXVECTOR3 vVehiclePos(matWorld._41, matWorld._42, matWorld._43);

#ifdef DRAW_VEHICLE_BOUNDINGSPHERE
			mDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, true );
			mDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCCOLOR);
			mDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCCOLOR);
#endif
			if( i != mCurrentVehicle && mCurrentCamera->FrustumSphereCheck( vVehiclePos, mVehicles[i]->GetMesh()->GetBoundingSphere()->radius ) )
			{
				objectDrawnCount++;
				mVehicles[i]->Render( mShapeDrawer );			

#ifdef DRAW_VEHICLE_BOUNDINGSPHERE
				mDevice->SetTransform( D3DTS_WORLD, &matWorld );
				mVehicles[i]->GetMesh()->GetBoundingSphereMesh()->DrawSubset( 0 );
#endif			
			}
			else if(i == mCurrentVehicle)
			{
				objectDrawnCount++;
				mVehicles[i]->Render( mShapeDrawer );			
			}

#ifdef DRAW_VEHICLE_BOUNDINGSPHERE
			mDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, false );
#endif			
		}

		AddTriangles( mShapeDrawer->GetNumFaces() );
		AddVertices( mShapeDrawer->GetNumVertices() );		
		sprintf_s( mDebugString, "Shoot speed = %.f Jedi force strength = %.f", mShootSpeed, mImpulseStrength );
		mDebugStrings.push_back(mDebugString);
	}

	virtual void UpdateClient( double time, float elapsedTime )							
	{ 
		// Move / update character controller
		if( mCharacterController && mCurrentCamera == &mCamera )			
		{
			if(GetKeyState( 'U' ) & 0x8000 )
				mForward = 0; 
			else
				mForward = 1; 

			if(GetKeyState( 'J' ) & 0x8000 )
				mBackward = 0; 		
			else
				mBackward = 1; 

			if(GetKeyState( 'H' ) & 0x8000 )
				mLeft = 0; 		
			else
				mLeft = 1; 

			if(GetKeyState( 'K' ) & 0x8000 )
				mRight = 0; 		
			else
				mRight = 1; 	

			// Bullet 2.74 code
			btTransform xform;
			xform = mCharacterGhostObject->getWorldTransform();	

			btVector3 forwardDir = xform.getBasis()[2];			
			btVector3 upDir = xform.getBasis()[1];
			btVector3 strafeDir = xform.getBasis()[0];						

			forwardDir.normalize ();
			upDir.normalize ();
			strafeDir.normalize ();

			btVector3 walkDirection = btVector3(0.0, 0.0, 0.0);
			btScalar walkVelocity = btScalar(1.1f) * 50.0f; // 4 km/h -> 1.1 m/s
			btScalar walkSpeed = walkVelocity * elapsedTime;

			if (mLeft)
				walkDirection += strafeDir;

			if (mRight)
				walkDirection -= strafeDir;

			if (mForward)
				walkDirection -= forwardDir;

			if (mBackward)
				walkDirection += forwardDir;	

			mCharacterController->setWalkDirection(walkDirection * walkSpeed);	

			// Need to apply rotations			
			btMatrix3x3 basis;				

			forwardDir.setX( (float)sin( mCurrentCamera->GetYawAngle() ));
			forwardDir.setY( 0.0f );
			forwardDir.setZ( (float)cos( mCurrentCamera->GetYawAngle() ));		

			strafeDir.setX( (float)cos( mCurrentCamera->GetYawAngle() ));
			strafeDir.setY( 0.0f );
			strafeDir.setZ( (float)-sin( mCurrentCamera->GetYawAngle() ));

			upDir = xform.getBasis()[1];
			forwardDir.normalize ();
			upDir.normalize ();
			strafeDir.normalize ();

			basis.setValue(strafeDir.x(), strafeDir.y(), strafeDir.z(),
				upDir.x(), upDir.y(), upDir.z(),
				forwardDir.x(), forwardDir.y(), forwardDir.z() );

			xform.setBasis( basis );
			mCharacterController->getGhostObject()->setWorldTransform( xform );			
		}

		// Update current selected vehicle position
		D3DXVECTOR3 vehiclePosition = ConvertVector( mVehicles[mCurrentVehicle]->GetWorldTransform().getOrigin() );
		D3DXVECTOR3 vehicleDirection = ConvertVector( mVehicles[mCurrentVehicle]->GetDirection() );

		mChaseCamera.SetChasePosition( vehiclePosition );		
		mChaseCamera.SetChaseDirection( -(*D3DXVec3Normalize(&vehicleDirection, &vehicleDirection)) );			

		mCurrentCamera->FrameMove( elapsedTime );

		// Put a limit to the number of spheres that can be shoot
		if( mShootRigidBodies.size() > 20 )
		{
			btCollisionObject *obj = mShootRigidBodies.front();
			btRigidBody* body = btRigidBody::upcast(obj);
			if (body && body->getMotionState())
			{
				delete body->getMotionState();
			}
			mDynamicsWorld->removeCollisionObject( obj );
			SAFE_DELETE( obj );			
			mShootRigidBodies.pop_front();
		}

		// Draw bullet profiler
		if( mDynamicsWorld && !mAppDlg.GetCheckBox( IDC_TOGGLEPAUSESIMULATION )->GetChecked() )
		{
			mDynamicsWorld->stepSimulation( elapsedTime );			

			mVehicles[mCurrentVehicle]->Update( elapsedTime );			

			static double time_since_reset = 0.f;
			time_since_reset = CProfileManager::Get_Time_Since_Reset();

			char blockTime[128];
			double totalTime = 0;
			int frames_since_reset = CProfileManager::Get_Frame_Count_Since_Reset();
			mProfileIterator->First();		

			double parent_time = mProfileIterator->Is_Root() ? time_since_reset : mProfileIterator->Get_Current_Parent_Total_Time();

			double accumulated_time = 0.f;

			for (int i = 0; !mProfileIterator->Is_Done(); mProfileIterator->Next())
			{
				double current_total_time = mProfileIterator->Get_Current_Total_Time();
				accumulated_time += current_total_time;
				double fraction = parent_time > SIMD_EPSILON ? (current_total_time / parent_time) * 100 : 0.f;

				sprintf_s(blockTime,"%d -- %s (%.2f %%) :: %.3f ms / frame (%d calls)",
					++i, mProfileIterator->Get_Current_Name(), fraction,
					(current_total_time / (double)frames_since_reset),mProfileIterator->Get_Current_Total_Calls());				

				mDebugStrings.push_back(blockTime);			

				totalTime += current_total_time;
			}

			{
				sprintf_s(blockTime,"%s (%.3f %%) :: %.3f ms", "Unaccounted",				
					parent_time > SIMD_EPSILON ? ((parent_time - accumulated_time) / parent_time) * 100 : 0.f, parent_time - accumulated_time);

				mDebugStrings.push_back(blockTime);			
			}

			sprintf_s( mDebugString, "# objects = %d # pairs = %d", mDynamicsWorld->getNumCollisionObjects(), 
				mDynamicsWorld->getBroadphase()->getOverlappingPairCache()->getNumOverlappingPairs());
			mDebugStrings.push_back(mDebugString);
		}
	}

	void CycleVehicles()
	{
		mCurrentVehicle++;
		if( (size_t)mCurrentVehicle >= mVehicles.size() )
			mCurrentVehicle = 0;
	}

	void ChangeCamera()
	{
		if( mCurrentCamera == &mCamera )
			mCurrentCamera = &mChaseCamera;
		else
			mCurrentCamera = &mCamera;
	}

	virtual void GUIEventClient( UINT nEvent, int nControlID, CDXUTControl* pControl )	
	{ 
		switch( nControlID )
		{
		case IDC_SHOOTSPEED:	
			mShootSpeed = (float)mAppDlg.GetSlider( IDC_SHOOTSPEED )->GetValue();
			break;
		case IDC_JEDIFORCEIMPULSE:
			mImpulseStrength = (float)mAppDlg.GetSlider( IDC_JEDIFORCEIMPULSE )->GetValue();
			break;
		case IDC_JEDIMODE:
			mJediMode = (bool)mAppDlg.GetCheckBox( IDC_JEDIMODE )->GetChecked();
			break;
		case IDC_CYCLEVEHICLE:
			CycleVehicles();
			break;
		case IDC_CHANGECAMERA:
			ChangeCamera();
			break;
		case IDC_RESETSCENE:
			ResetScene();
			break;
		}

	}
	virtual void KeyboardProcClient( UINT nChar, bool bKeyDown, bool bAltDown )			
	{ 
		if( bKeyDown )
		{
			switch( nChar )
			{
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
				mProfileIterator->Enter_Child(nChar - 49); // 48 is ascii decimal equivalent for character '0'
				break;
			case '0':
				mProfileIterator->Enter_Parent();
				break;
			case VK_SPACE:
				ResetScene();
				break;
			case 'I':
			case 'i':
				mAppDlg.GetCheckBox( IDC_TOGGLEPAUSESIMULATION )->SetChecked( !mAppDlg.GetCheckBox( IDC_TOGGLEPAUSESIMULATION )->GetChecked() );
				break;		
			case VK_F4:
				ChangeCamera();
				break;
			case VK_F5:
				CycleVehicles();
				break;
			}
		}
	}
	virtual LRESULT MsgProcClient( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )	
	{ 
		if( mCurrentCamera == &mCamera)
			mCamera.HandleMessages( hWnd, uMsg, wParam, lParam );

		return 0; 
	}
	virtual void MouseProcClient( bool bLeftButtonDown, bool bRightButtonDown, bool bMiddleButtonDown, bool bSideButton1Down, bool bSideButton2Down, int nMouseWheelDelta, int xPos, int yPos ) 
	{ 
		if( mCurrentCamera != &mCamera )
			return;

		btVector3 rayTo = ConvertVector(GetRayTo( xPos, yPos ));	

		if( bRightButtonDown && !mJediMode )
		{
			btVector3 origin = ConvertVector( mCamera.GetEyePt() );
			btVector3 destination = rayTo;

			btVector3 linVel(destination[0] - origin[0], destination[1]- origin[1], destination[2]- origin[2] );
			linVel.normalize();
			linVel *= mShootSpeed;

			btTransform startTransform;
			startTransform.setIdentity();		
			startTransform.setOrigin( origin );		
			btRigidBody *pBody =  CreateRigidBody( 10.0f, startTransform, mShootShape );
			mShootRigidBodies.push_back( pBody );
			pBody->getWorldTransform().setOrigin(origin);
			pBody->getWorldTransform().setRotation(btQuaternion(0,0,0,1));
			pBody->setLinearVelocity(linVel);
			pBody->setAngularVelocity(btVector3(0,0,0));
			pBody->setCcdMotionThreshold(1);
			pBody->setCcdSweptSphereRadius(0.2f);
			mDynamicsWorld->addRigidBody( pBody );
		}

		if( bRightButtonDown && mJediMode )
		{
			// Fire
			btRigidBody *pBody = NULL;

			btVector3 pos = ConvertVector(mCamera.GetEyePt());			

			btCollisionWorld::ClosestRayResultCallback rayCallback(pos,rayTo);
			mDynamicsWorld->rayTest(pos, rayTo, rayCallback);
			if (rayCallback.hasHit())
			{
				btRigidBody* body = btRigidBody::upcast(rayCallback.m_collisionObject);

				// Other exclusions?
				if ( body && !(body->isStaticObject() || body->isKinematicObject()) )
				{	
					// Move ghost object to world space clicked by user
					btTransform trn = mGhostObject->getWorldTransform();					
					trn.setOrigin(rayCallback.m_hitPointWorld);
					mGhostObject->setWorldTransform(trn);															
				}
			}
		}	
		else if( mJediMode )
		{
			rayTo.normalize();
			rayTo *= mImpulseStrength;

			for(int i = 0; i < mGhostObject->getNumOverlappingObjects(); i++)
			{
				// Upcast to find out if its a rigid body				
				btRigidBody* rigidBody = btRigidBody::upcast(mGhostObject->getOverlappingObject(i));
				if(rigidBody)
				{
					rigidBody->activate(true);
					rigidBody->applyCentralImpulse(rayTo);
				}
			}
			// Reset position
			btTransform trn = mGhostObject->getWorldTransform();					
			trn.setOrigin(btVector3(0.0f, 10000.0f, 0.0f));
			mGhostObject->setWorldTransform(trn);			
		}
	}

public:
	BulletRigidBodies()
	{
		mDynamicsWorld		= NULL;
		mBroadPhase			= NULL;
		mConstraintSolver	= NULL;
		mCollisionDispatcher	= NULL;
		mCollisionConfiguration	= NULL;

		mShapeDrawer		= NULL;

		mSceneMesh			= NULL;		
		mTriMeshData		= NULL;	
		mHeightMap			= NULL;

		mCurrentVehicle		= 0;
		mShootSpeed			= 400.f;
		mImpulseStrength	= 400.f;
		mJediMode			= false;
		mIsGhostObject		= true;

		mCharacterController = NULL;
		mCharacterGhostObject = NULL;		
		mForward	= 0;
		mBackward	= 0;
		mLeft		= 0;
		mRight		= 0;
		mJump		= 0;

		mProfileIterator = CProfileManager::Get_Iterator();
	}

	~BulletRigidBodies()
	{
		SAFE_DELETE( mHeightMap );		

		SAFE_DELETE( mShapeDrawer);
		SAFE_DELETE( mSceneMesh );		

		SAFE_DELETE( mDynamicsWorld );
		SAFE_DELETE( mConstraintSolver );
		SAFE_DELETE( mBroadPhase );
		SAFE_DELETE( mCollisionDispatcher );
		SAFE_DELETE( mCollisionConfiguration );		

		CProfileManager::Release_Iterator(mProfileIterator);
	}

private:
	// Rudimentary bullet variables
	btDynamicsWorld						*mDynamicsWorld;
	btDefaultCollisionConfiguration		*mCollisionConfiguration;
	btSequentialImpulseConstraintSolver	*mConstraintSolver;
	btCollisionDispatcher				*mCollisionDispatcher;	
	btBroadphaseInterface				*mBroadPhase;
	btAlignedObjectArray<btCollisionShape*> mCollisionShapes;

	// Use for shooting sphere
	btCollisionShape						*mShootShape;	
	std::list< btRigidBody *>				mShootRigidBodies;	
	float			mShootSpeed;
	float			mImpulseStrength;
	bool			mJediMode;

	// Ghost object for the jedi force
	btGhostPairCallback *mGhostPairCallback; 
	btGhostObject		*mGhostObject;
	bool				mIsGhostObject; // Variable for the user pointer function, because we don't want to draw this object

	// Character controller
	btPairCachingGhostObject*			mCharacterGhostObject;
	btKinematicCharacterController*		mCharacterController;
	int mForward;
	int mBackward;
	int mLeft;
	int mRight;
	int mJump;

	// A DirectX shape drawer class
	ShapeDrawer			*mShapeDrawer;

	// Ragdolls and vehicles
	std::vector<RagDoll *> mRagDolls;
	std::vector<Vehicle *> mVehicles;
	int					   mCurrentVehicle;
	ChaseCamera			   mChaseCamera;
	Camera				  *mCurrentCamera;

	// Variable to hold the static scene
	Mesh			*mSceneMesh;
	TriMeshData		*mTriMeshData;		
	HeightMap		*mHeightMap;	

	// Bullet profiler - take note, it does not profile any rendering calls
	CProfileIterator* mProfileIterator;
};

//--------------------------------------------------------------------------------------
// Entry point to the program. Initializes everything and goes into a message processing 
// loop. Idle time is used to render the scene.
//--------------------------------------------------------------------------------------
int WINAPI wWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nCmdShow )
{
	// Enable run-time memory check for debug builds.
#if defined(DEBUG) | defined(_DEBUG)
	_CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif

	LogOSAndAdapter();

	BulletRigidBodies *app = new BulletRigidBodies();				
	return Initialize( 1024, 768, L"Bullet Rigid Body Physics", app );	
}