#include "dxut.h"
#include "ShapeDrawer.h"




#include "Vertex.h"
#include "LightMaterial.h"
#include "Utilities.h"
#include "BulletUtilities.h"

ShapeDrawer::ShapeDrawer(void)
{
	LOG_DEBUG( "ShapeDrawer constructor called." );
	mBoxMesh			= NULL;
	mSphereMesh			= NULL;
	mCylinderMesh		= NULL;
	mConeMesh			= NULL;		
}

ShapeDrawer::~ShapeDrawer(void)
{
	LOG_DEBUG( "ShapeDrawer destructor called." );
	Destroy();
}

void ShapeDrawer::Render(const btCollisionShape* shape, btTransform &trn)
{
	int shapetype = shape->getShapeType();	
	switch (shapetype)
	{	
		case COMPOUND_SHAPE_PROXYTYPE:
		{
			const btCompoundShape* compoundShape = static_cast<const btCompoundShape*>(shape);
			for (int i = compoundShape->getNumChildShapes()-1; i >= 0; i--)
			{
				btTransform childTrans = compoundShape->getChildTransform(i);
				const btCollisionShape* colShape = compoundShape->getChildShape(i);				
				Render(colShape, trn * childTrans);
			}
		}
		break;
	case BOX_SHAPE_PROXYTYPE:
		{
			btVector3 halfExtent = ((btBoxShape*)shape)->getHalfExtentsWithMargin();

			D3DXMATRIX matWorld;
			D3DXMATRIX matScale, matRot, matTrans;					
			D3DXMatrixScaling( &matScale, 2*halfExtent[0], 2*halfExtent[1], 2*halfExtent[2] );
			D3DXMatrixTranslation( &matTrans, trn.getOrigin().x(), trn.getOrigin().y(), trn.getOrigin().z());
			D3DXMatrixRotationQuaternion( &matRot, &D3DXQUATERNION(
				trn.getRotation().x(),
				trn.getRotation().y(),
				trn.getRotation().z(),
				trn.getRotation().w()));

			D3DXMatrixMultiply(&matWorld, &matScale, &matRot);
			D3DXMatrixMultiply(&matWorld, &matWorld, &matTrans);
			
			mDevice->SetTransform( D3DTS_WORLD, &matWorld );			
			mBoxMesh->DrawSubset( 0 );

			mNumVertices += mBoxMesh->GetNumVertices();
			mNumFaces	+= mBoxMesh->GetNumFaces();
			mNumObjects++;
		}
		break;
	case CYLINDER_SHAPE_PROXYTYPE:
		{
			int upAxis = ((btCylinderShape*)shape)->getUpAxis();
			float radius =  ((btCylinderShape*)shape)->getRadius();
			float halfHeight =  ((btCylinderShape*)shape)->getHalfExtentsWithMargin()[upAxis];

			D3DXMATRIX matWorld;
			D3DXMATRIX matScale, matRot, matTrans;					
			D3DXMatrixScaling( &matScale, radius, halfHeight, radius );
			D3DXMatrixTranslation( &matTrans, trn.getOrigin().x(), trn.getOrigin().y(), trn.getOrigin().z());

			D3DXMatrixRotationQuaternion( &matRot, &D3DXQUATERNION(
				trn.getRotation().x(),
				trn.getRotation().y(),
				trn.getRotation().z(),
				trn.getRotation().w()));		

			D3DXMatrixMultiply(&matWorld, &matScale, &matRot);
			D3DXMatrixMultiply(&matWorld, &matWorld, &matTrans);

			D3DXMatrixRotationX( &matRot, D3DXToRadian(-90.0f));
			D3DXMatrixMultiply(&matWorld, &matRot, &matWorld);

			mDevice->SetTransform( D3DTS_WORLD, &matWorld );							
			mCylinderMesh->DrawSubset( 0 );

			mNumVertices += mCylinderMesh->GetNumVertices();
			mNumFaces	+= mCylinderMesh->GetNumFaces();
			mNumObjects++;
		}
		break;		
		case CONE_SHAPE_PROXYTYPE:
		{
			float radius =  ((btConeShape*)shape)->getRadius();
			float halfHeight =  ((btConeShape*)shape)->getHeight();

			D3DXMATRIX matWorld;
			D3DXMATRIX matScale, matRot, matTrans;					
			D3DXMatrixScaling( &matScale, radius, halfHeight, radius );
			D3DXMatrixTranslation( &matTrans, trn.getOrigin().x(), trn.getOrigin().y(), trn.getOrigin().z());

			D3DXMatrixRotationQuaternion( &matRot, &D3DXQUATERNION(
			trn.getRotation().x(),
			trn.getRotation().y(),
			trn.getRotation().z(),
			trn.getRotation().w()));		
	
			D3DXMatrixMultiply(&matWorld, &matScale, &matRot);
			D3DXMatrixMultiply(&matWorld, &matWorld, &matTrans);

			D3DXMatrixRotationX( &matRot, D3DXToRadian(-90.0f));
			D3DXMatrixMultiply(&matWorld, &matRot, &matWorld);

			mDevice->SetTransform( D3DTS_WORLD, &matWorld );
			mConeMesh->DrawSubset( 0 );

			mNumVertices += mConeMesh->GetNumVertices();
			mNumFaces	+= mConeMesh->GetNumFaces();
			mNumObjects++;
		}
		break;		
	case SPHERE_SHAPE_PROXYTYPE:
		{
			float fRadius = ((btSphereShape*)shape)->getMargin();			

			D3DXMATRIX matWorld;
			D3DXMATRIX matScale, matRot, matTrans;					
			D3DXMatrixScaling( &matScale, fRadius, fRadius, fRadius );
			D3DXMatrixTranslation( &matTrans, trn.getOrigin().x(), trn.getOrigin().y(), trn.getOrigin().z());
			D3DXMatrixRotationQuaternion( &matRot, &D3DXQUATERNION(
				trn.getRotation().x(),
				trn.getRotation().y(),
				trn.getRotation().z(),
				trn.getRotation().w()));

			D3DXMatrixMultiply(&matWorld, &matScale, &matRot);
			D3DXMatrixMultiply(&matWorld, &matWorld, &matTrans);			

			mDevice->SetTransform( D3DTS_WORLD, &matWorld );	
			mSphereMesh->DrawSubset( 0 );

			mNumVertices += mSphereMesh->GetNumVertices();
			mNumFaces	+= mSphereMesh->GetNumFaces();
			mNumObjects++;
		}
		break;
	default:
		{
			if( shape->isConvex() )
			{
				ShapeCache*	sc = Cache((btConvexShape*)shape);

				//btShapeHull* hull = &sc->shapeHull;
				btShapeHull* hull = (btShapeHull*)shape->getUserPointer();

				if ( hull )
				{
					D3DXMATRIX matWorld = ConvertMatrix( trn );
					//D3DXMatrixMultiply(&matWorld, &matWorld, &mMatCompound);			
					mDevice->SetTransform( D3DTS_WORLD, &matWorld );	
					mDevice->SetFVF(D3DFVF_VERTEXNORMAL);

					int numTri =  hull->numTriangles();

					if( numTri > 0 )
					{
						int index = 0;
						const unsigned int* idx = hull->getIndexPointer();
						const btVector3* vtx = hull->getVertexPointer();						

						VertexNormal *pVertex = new VertexNormal[numTri * 3];

						for (int i = 0; i < hull->numTriangles (); i++)
						{
							int i1 = index++;
							int i2 = index++;
							int i3 = index++;
							btAssert(i1 < hull->numIndices () &&
								i2 < hull->numIndices () &&
								i3 < hull->numIndices ());

							int index1 = idx[i1];
							int index2 = idx[i2];
							int index3 = idx[i3];
							btAssert(index1 < hull->numVertices () &&
								index2 < hull->numVertices () &&
								index3 < hull->numVertices ());

							btVector3 v1 = vtx[index1];
							btVector3 v2 = vtx[index2];
							btVector3 v3 = vtx[index3];
							btVector3 normal = (v3-v1).cross(v1-v2);
							normal.normalize ();

							pVertex[i1].position = D3DXVECTOR3(v1.x(), v1.y(), v1.z());
							pVertex[i2].position = D3DXVECTOR3(v2.x(), v2.y(), v2.z());
							pVertex[i3].position = D3DXVECTOR3(v3.x(), v3.y(), v3.z());

							pVertex[i1].normal = D3DXVECTOR3( normal.x(), normal.y(), normal.z() );
							pVertex[i2].normal = D3DXVECTOR3( normal.x(), normal.y(), normal.z() );
							pVertex[i3].normal = D3DXVECTOR3( normal.x(), normal.y(), normal.z() );
						}
						mDevice->DrawPrimitiveUP(D3DPT_TRIANGLELIST, numTri, (LPVOID) pVertex, sizeof(VertexNormal));
						delete [] pVertex;

						mNumVertices += hull->numVertices();
						mNumFaces	+= hull->numTriangles();
						mNumObjects++;
					}
				}
			}			
		}
		break;
	}
}

void ShapeDrawer::RenderConcaveShapes( D3DXMATRIX *world, size_t index )
{	
	if( index < mConcaveShapeCaches.size() && mConcaveShapeCaches[index] )
	{
		mDevice->SetTransform( D3DTS_WORLD, world);		
		// the concave shapes only have vertices information
		mDevice->SetStreamSource( 0, mConcaveShapeCaches[index]->vertexBuffer, 0, sizeof(VertexNormal) );
		mDevice->SetFVF( D3DFVF_VERTEXNORMAL );
		mDevice->DrawPrimitive( D3DPT_TRIANGLELIST, 0,  mConcaveShapeCaches[index]->numFaces );

		mNumVertices += mConcaveShapeCaches[index]->numFaces * 3;
		mNumFaces	+= mConcaveShapeCaches[index]->numFaces;
		mNumObjects++;
	}
}

void ShapeDrawer::Create( LPDIRECT3DDEVICE9 device )
{
	mDevice = device;
	D3DXCreateBox( mDevice, 1.0f, 1.0f, 1.0f, &mBoxMesh, NULL );
	D3DXCreateSphere( mDevice, 1.0f, 20, 20, &mSphereMesh, NULL );	
	D3DXCreateCylinder( mDevice, 1.0f, 1.0f, 2.0f, 15, 10, &mCylinderMesh, NULL );	
	D3DXCreateCylinder( mDevice, 1.0f, 0.0f, 1.0f, 15, 10, &mConeMesh, NULL );	
}

void ShapeDrawer::Reset( LPDIRECT3DDEVICE9 device )
{	
}

void ShapeDrawer::Lost()
{
}

void ShapeDrawer::Destroy()
{
	SAFE_RELEASE( mBoxMesh );
	SAFE_RELEASE( mSphereMesh );
	SAFE_RELEASE( mCylinderMesh );
	SAFE_RELEASE( mConeMesh );	

	for(size_t i = 0; i < mConcaveShapeCaches.size(); i++)
	{
		SAFE_RELEASE( mConcaveShapeCaches[i]->vertexBuffer );
		SAFE_DELETE( mConcaveShapeCaches[i] );
	}
	mConcaveShapeCaches.clear();	
	for (int i = 0;i < mShapeCaches.size(); i++)
	{
		mShapeCaches[i]->~ShapeCache();
		btAlignedFree(mShapeCaches[i]);
	}	
	mShapeCaches.clear();
}

void ShapeDrawer::AddConcaveShape( btConcaveShape *shape )
{
	ShapeDrawer::ConcaveShapeCache *cache = new ConcaveShapeCache();

	btVector3 worldMin( -10000, -10000, -10000 );
	btVector3 worldMax( 10000, 10000, 10000 );
	DxDrawcallback drawCallback;	

	// Generates a disjointed triangle for each primitive
	shape->processAllTriangles(&drawCallback, worldMin, worldMax);	
	cache->numFaces = drawCallback.tri;	

	int size = drawCallback.verts * sizeof(VertexNormal);
	mDevice->CreateVertexBuffer( size, 0, D3DFVF_VERTEXNORMAL, D3DPOOL_MANAGED, &cache->vertexBuffer, NULL );

	VOID* pVertices;
	cache->vertexBuffer->Lock( 0, size, (void**)&pVertices, 0 );	
	memcpy( pVertices, &drawCallback.vertices[0], size);
	cache->vertexBuffer->Unlock();

	mConcaveShapeCaches.push_back( cache );
}

ShapeDrawer::ShapeCache* ShapeDrawer::Cache(btConvexShape* shape)
{
	ShapeCache*		sc=(ShapeCache*)shape->getUserPointer();
	if(!sc)
	{
		sc=new(btAlignedAlloc(sizeof(ShapeCache),16)) ShapeCache(shape);
		sc->shapeHull.buildHull(shape->getMargin());
		mShapeCaches.push_back(sc);
		shape->setUserPointer(sc);
		/* Build edges	*/ 
		const int			ni=sc->shapeHull.numIndices();
		const int			nv=sc->shapeHull.numVertices();
		const unsigned int*	pi=sc->shapeHull.getIndexPointer();
		const btVector3*	pv=sc->shapeHull.getVertexPointer();
		btAlignedObjectArray<ShapeCache::Edge*>	edges;
		sc->edges.reserve(ni);
		edges.resize(nv*nv,0);
		for(int i=0;i<ni;i+=3)
		{
			const unsigned int* ti=pi+i;
			const btVector3		nrm=cross(pv[ti[1]]-pv[ti[0]],pv[ti[2]]-pv[ti[0]]).normalized();
			for(int j=2,k=0;k<3;j=k++)
			{
				const unsigned int	a=ti[j];
				const unsigned int	b=ti[k];
				ShapeCache::Edge*&	e=edges[btMin(a,b)*nv+btMax(a,b)];
				if(!e)
				{
					sc->edges.push_back(ShapeCache::Edge());
					e=&sc->edges[sc->edges.size()-1];
					e->n[0]=nrm;e->n[1]=-nrm;
					e->v[0]=a;e->v[1]=b;
				}
				else
				{
					e->n[1]=nrm;
				}
			}
		}
	}
	return(sc);
}