#include "dxut.h"
#include "Vehicle.h"
#include "Mesh.h"
#include "BulletUtilities.h"
#include "ShapeDrawer.h"

//#define DEFAULT_VEHICLE_SETUP
//#define DEBUG_VEHICLE_WHEEL

Vehicle::Vehicle(void)
{
	LOG_DEBUG( "Vehicle constructor called." );

	mChassisShape = NULL;
	mCompoundShape = NULL;
	mVehicleRaycaster = NULL;
	mRaycastVehicle = NULL;
	mWheelDebugMesh = NULL;
	mChassisMesh = NULL;
	mWheelMesh = NULL;

	mSteeringValue = 0.f;
	mMaxEngineForce = 3000.f; // This should be engine/velocity dependent
	mMaxBreakingForce = 100.f;
	mSteeringIncrement = 0.04f;
	mSteeringClamp = 0.4f;
	mEngineForce = 0.f;
	mBreakingForce = 0.f;
	mSuspensionStiffness = 20.f;
	mSuspensionDamping = 2.3f;
	mSuspensionCompression = 4.4f;
	mWheelFriction = 1000;//1e30f;
	mRollInfluence = 0.1f;//1.0f;
	mChassisMesh = NULL;
	mWheelMesh	= NULL;
	mWheelDebugMesh = NULL;

#ifdef DEFAULT_VEHICLE_SETUP
	mWheelWidth = 0.4f;
	mWheelRadius = 0.5f;
#else
	mWheelWidth = 0.4f;
	mWheelRadius = 1.f;
#endif
	mSize = 1.0f;
}

Vehicle::~Vehicle(void)
{
	LOG_DEBUG( "Vehicle destructor Called." );
	Destroy();
}

void Vehicle::Destroy()
{
	mDynamicsWorld->removeVehicle( mRaycastVehicle );
	SAFE_DELETE( mCompoundShape );
	SAFE_DELETE( mChassisShape );
	SAFE_DELETE( mVehicleRaycaster );
	SAFE_DELETE( mRaycastVehicle );
	SAFE_RELEASE( mWheelDebugMesh );
	SAFE_DELETE( mChassisMesh );
	SAFE_DELETE( mWheelMesh );
	SAFE_DELETE( mSimplifiedConvexShape );
}

bool Vehicle::Create( LPDIRECT3DDEVICE9 device, btDynamicsWorld *dynamicsWorld, btVector3 &position, btScalar mass )
{
	mDevice = device;
	mDynamicsWorld = dynamicsWorld;

	mChassisMesh = new Mesh();
	mChassisMesh->Create( mDevice, "Media//chassis.x" );	

	mWheelMesh = new Mesh();
	mWheelMesh->Create( mDevice, "Media//wheel.x" );	

#ifdef DEFAULT_VEHICLE_SETUP
	// create a chassis shape that is proportfion to the size	
	mChassisShape	 = new btBoxShape(btVector3(1.f * mSize, 0.5f, 2.0f * mSize));
#else
	mSize = 1.7f;

	mChassisShape = new btConvexHullShape();

	// Extract vertices from vehicle mesh
	Vertex *v = 0;		
	mChassisMesh->GetMesh()->LockVertexBuffer(0, (void**)&v);
	for( DWORD i = 0; i < mChassisMesh->GetMesh()->GetNumVertices(); i++ )
	{
		((btConvexHullShape*)mChassisShape)->addPoint( btVector3( v[i].position.x, v[i].position.y, v[i].position.z ) );
	}
	mChassisMesh->GetMesh()->UnlockVertexBuffer();
#endif

	mSimplifiedConvexShape = CreateConvexHullShape( mChassisMesh->GetMesh() );

	mCompoundShape = new btCompoundShape();
	//mWheelWidth = 0.4f * mSize;
	//mWheelRadius = 0.5f * mSize;

	btTransform localTrans;
	localTrans.setIdentity();
	localTrans.setOrigin( btVector3(0.0f, 1.0f, 0.0f ) );

#ifdef DEFAULT_VEHICLE_SETUP
	mCompoundShape->addChildShape(localTrans, mChassisShape);
#else
	mCompoundShape->addChildShape(localTrans, mSimplifiedConvexShape);
#endif

	localTrans.setOrigin( position );	
	mChassisRigidBody = CreateRigidBody( mass, localTrans, mCompoundShape );
	mChassisRigidBody->setCenterOfMassTransform( localTrans );
	mChassisRigidBody->setLinearVelocity(btVector3(0,0,0));
	mChassisRigidBody->setAngularVelocity(btVector3(0,0,0));	
	mDynamicsWorld->addRigidBody(mChassisRigidBody );

	mDynamicsWorld->getBroadphase()->getOverlappingPairCache()->cleanProxyFromPairs(mChassisRigidBody->getBroadphaseHandle(), mDynamicsWorld->getDispatcher());

	int rightIndex = 0;
	int upIndex = 1;
	int forwardIndex = 2;
	btVector3 wheelDirectionCS0(0,-1,0);
	btVector3 wheelAxleCS(-1,0,0);
	btScalar suspensionRestLength(0.6);

	mVehicleRaycaster = new btDefaultVehicleRaycaster(mDynamicsWorld);
	mRaycastVehicle = new btRaycastVehicle(mVehicleTuning, mChassisRigidBody, mVehicleRaycaster);				

	// Never deactivate the vehicle
	mChassisRigidBody->setActivationState(DISABLE_DEACTIVATION);

	mDynamicsWorld->addVehicle(mRaycastVehicle);

#ifdef DEFAULT_VEHICLE_SETUP
	float connectionHeight = 1.2f;	
#else
	float connectionHeight = 1.9f;	
#endif

	bool isFrontWheel = true;		
	//choose coordinate system
	mRaycastVehicle->setCoordinateSystem(rightIndex, upIndex, forwardIndex);

	btVector3 connectionPointCS0(mSize-(.3f*mWheelWidth),connectionHeight,2.f*mSize-mWheelRadius);		
	mRaycastVehicle->addWheel(connectionPointCS0,wheelDirectionCS0,wheelAxleCS,suspensionRestLength,mWheelRadius,mVehicleTuning,isFrontWheel);

	connectionPointCS0 = btVector3(-mSize+(.3f*mWheelWidth),connectionHeight,2.f*mSize-mWheelRadius);
	mRaycastVehicle->addWheel(connectionPointCS0,wheelDirectionCS0,wheelAxleCS,suspensionRestLength,mWheelRadius,mVehicleTuning,isFrontWheel);

	connectionPointCS0 = btVector3(-mSize+(.3f*mWheelWidth),connectionHeight,-2.f*mSize+mWheelRadius);
	isFrontWheel = false;
	mRaycastVehicle->addWheel(connectionPointCS0,wheelDirectionCS0,wheelAxleCS,suspensionRestLength,mWheelRadius,mVehicleTuning,isFrontWheel);

	connectionPointCS0 = btVector3(mSize-(.3f*mWheelWidth),connectionHeight,-2.f*mSize+mWheelRadius);
	mRaycastVehicle->addWheel(connectionPointCS0,wheelDirectionCS0,wheelAxleCS,suspensionRestLength,mWheelRadius,mVehicleTuning,isFrontWheel);

	for (int i=0;i<mRaycastVehicle->getNumWheels();i++)
	{
		btWheelInfo& wheel = mRaycastVehicle->getWheelInfo(i);
		wheel.m_suspensionStiffness = mSuspensionStiffness;
		wheel.m_wheelsDampingRelaxation = mSuspensionDamping;
		wheel.m_wheelsDampingCompression = mSuspensionCompression;
		wheel.m_frictionSlip = mWheelFriction;
		wheel.m_rollInfluence = mRollInfluence;			
	}

	// Create a wheel shape
#ifdef DEBUG_VEHICLE_WHEEL
	btCylinderShapeX wheelShape(btVector3(mWheelWidth,mWheelRadius,mWheelRadius));
	int upAxis = ((btCylinderShapeX)wheelShape).getUpAxis();
	float radius =  ((btCylinderShapeX)wheelShape).getRadius();
	float halfHeight = ((btCylinderShapeX)wheelShape).getHalfExtentsWithMargin()[upAxis];		

	D3DXCreateCylinder( mDevice, radius, radius, halfHeight* 2.0f, 15, 10, &mWheelDebugMesh, NULL );
#endif 

	return true;
}

void Vehicle::Update( float elapsedTime )
{
	if(GetKeyState( VK_UP ) & 0x8000 )
	{
		mEngineForce = mMaxEngineForce;
		mBreakingForce = 0.f;			
	}
	else
	{
		mEngineForce = 0.f;
	}		

	if(GetKeyState( VK_DOWN ) & 0x8000 )
	{
		mEngineForce = -mMaxEngineForce;
		mBreakingForce = 0.f;
	}		
	else
	{
		mBreakingForce = 0.f; 			
	}  		

	if(GetKeyState( VK_LEFT ) & 0x8000 )		
	{
		mSteeringValue -= mSteeringIncrement;
		if ( mSteeringValue < -mSteeringClamp)
			mSteeringValue = -mSteeringClamp;
	}
	else
	{
		// Return wheel to original position
		if ( mSteeringValue <= 0.0f )
			mSteeringValue += mSteeringIncrement;
	}

	if(GetKeyState( VK_RIGHT ) & 0x8000 )			
	{
		mSteeringValue += mSteeringIncrement;
		if(mSteeringValue > mSteeringClamp)
			mSteeringValue = mSteeringClamp;
	}	
	else
	{
		// Return wheel to original position
		if( mSteeringValue >= 0.0f )
			mSteeringValue -= mSteeringIncrement;			
	}

	int wheelIndex = 2;
	mRaycastVehicle->applyEngineForce(mEngineForce,wheelIndex);
	mRaycastVehicle->setBrake(mBreakingForce,wheelIndex);
	wheelIndex = 3;
	mRaycastVehicle->applyEngineForce(mEngineForce,wheelIndex);
	mRaycastVehicle->setBrake(mBreakingForce,wheelIndex);

	wheelIndex = 0;
	mRaycastVehicle->setSteeringValue(mSteeringValue,wheelIndex);
	wheelIndex = 1;
	mRaycastVehicle->setSteeringValue(mSteeringValue,wheelIndex);		

	// Update vehicle direction
	btWheelInfo& wheel = mRaycastVehicle->getWheelInfo( 0 );
	btVector3 up = -wheel.m_raycastInfo.m_wheelDirectionWS;
	const btVector3& right = wheel.m_raycastInfo.m_wheelAxleWS;
	mForward = up.cross(right);
	mForward.normalize();	
}

void Vehicle::Render( ShapeDrawer *shapeDrawer )
{	
	// Draw wheels (cylinders)
	for (int i=0;i<mRaycastVehicle->getNumWheels();i++)
	{
		//synchronize the wheels with the (interpolated) chassis worldtransform
		mRaycastVehicle->updateWheelTransform(i, true );				
		D3DXMATRIX matWorld = ConvertMatrix( mRaycastVehicle->getWheelInfo(i).m_worldTransform );				
		btVector3 pos = mRaycastVehicle->getWheelInfo(i).m_worldTransform.getOrigin();

#ifndef DEFAULT_VEHICLE_SETUP
		mDevice->SetTransform( D3DTS_WORLD, &matWorld );
		if( mWheelMesh )
		{
			mWheelMesh->GetMesh()->DrawSubset( 0 );
		}
#endif

#ifdef DEBUG_VEHICLE_WHEEL
		D3DXMATRIX matRot;					
		D3DXMatrixRotationY( &matRot, D3DXToRadian(-90.0f));
		D3DXMatrixMultiply(&matWorld, &matRot, &matWorld);
		mDevice->SetTransform( D3DTS_WORLD, &matWorld );
		mWheelDebugMesh->DrawSubset( 0 ) ;			
#endif		
	}

	// Draw chassis
	{
		btTransform trn;
		mRaycastVehicle->getRigidBody()->getMotionState()->getWorldTransform( trn );		

#ifndef DEFAULT_VEHICLE_SETUP

		if( mChassisMesh )
		{
			D3DXMATRIX matWorld = ConvertMatrix( trn );		
			D3DXMATRIX matRot, matTrans, matScale;
			// offset 1.0 for the wheel
			D3DXMatrixTranslation( &matTrans, 0.0f, 1.0f, 0.0f );
			D3DXMatrixMultiply(&matWorld, &matTrans, &matWorld);		

			mDevice->SetTransform( D3DTS_WORLD, &matWorld );		
			mChassisMesh->GetMesh()->DrawSubset( 0 );
		}
#else
		if( shapeDrawer )
		{
			shapeDrawer->Render( mRaycastVehicle->getRigidBody()->getCollisionShape(),  trn );						
		}
#endif			
	}
}


